@@ -4,7 +4,7 @@ import abc
 import collections
 import base64
 import functools
-from io import BytesIO
+import io
 import itertools
 import logging
 import math
@@ -12,19 +12,16 @@ from operator import itemgetter, attrgetter
 import pickle
 import platform
 import random
+import shutil
 import subprocess
 import sys
+import tempfile
 import textwrap
 import types
 import unittest
 import weakref
 import os
 
-try:
-    from StringIO import StringIO
-except ImportError:
-    from io import StringIO
-
 import pytest
 
 try:
@@ -69,16 +66,21 @@ def pickle_depickle(obj, protocol=cloudpickle.DEFAULT_PROTOCOL):
     return pickle.loads(cloudpickle.dumps(obj, protocol=protocol))
 
 
-class CloudPicklerTest(unittest.TestCase):
-    def setUp(self):
-        self.file_obj = StringIO()
-        self.cloudpickler = cloudpickle.CloudPickler(self.file_obj, 2)
+def _escape(raw_filepath):
+    # Ugly hack to embed filepaths in code templates for windows
+    return raw_filepath.replace("\\", r"\\\\")
 
 
 class CloudPickleTest(unittest.TestCase):
 
     protocol = cloudpickle.DEFAULT_PROTOCOL
 
+    def setUp(self):
+        self.tmpdir = tempfile.mkdtemp(prefix="tmp_cloudpickle_test_")
+
+    def tearDown(self):
+        shutil.rmtree(self.tmpdir)
+
     def test_itemgetter(self):
         d = range(10)
         getter = itemgetter(1)
@@ -113,7 +115,7 @@ class CloudPickleTest(unittest.TestCase):
     # Regression test for SPARK-3415
     def test_pickling_file_handles(self):
         out1 = sys.stderr
-        out2 = pickle.loads(cloudpickle.dumps(out1))
+        out2 = pickle.loads(cloudpickle.dumps(out1, protocol=self.protocol))
         self.assertEqual(out1, out2)
 
     def test_func_globals(self):
@@ -124,7 +126,8 @@ class CloudPickleTest(unittest.TestCase):
         global exit
         exit = Unpicklable()
 
-        self.assertRaises(Exception, lambda: cloudpickle.dumps(exit))
+        self.assertRaises(Exception, lambda: cloudpickle.dumps(
+            exit, protocol=self.protocol))
 
         def foo():
             sys.exit(0)
@@ -166,7 +169,8 @@ class CloudPickleTest(unittest.TestCase):
                          buffer_obj.tobytes())
 
     def test_lambda(self):
-        self.assertEqual(pickle_depickle(lambda: 1)(), 1)
+        self.assertEqual(
+                pickle_depickle(lambda: 1, protocol=self.protocol)(), 1)
 
     def test_nested_lambdas(self):
         a, b = 1, 2
@@ -185,10 +189,10 @@ class CloudPickleTest(unittest.TestCase):
                 return base if n <= 1 else n * g(n - 1)
             return g
 
-        g1 = pickle_depickle(f1())
+        g1 = pickle_depickle(f1(), protocol=self.protocol)
         self.assertEqual(g1(), g1)
 
-        g2 = pickle_depickle(f2(2))
+        g2 = pickle_depickle(f2(2), protocol=self.protocol)
         self.assertEqual(g2(5), 240)
 
     def test_closure_none_is_preserved(self):
@@ -235,7 +239,7 @@ class CloudPickleTest(unittest.TestCase):
 
             return g
 
-        g = pickle_depickle(f())
+        g = pickle_depickle(f(), protocol=self.protocol)
         self.assertEqual(g(), 2)
 
     def test_dynamically_generated_class_that_uses_super(self):
@@ -345,24 +349,27 @@ class CloudPickleTest(unittest.TestCase):
 
         if np:
             # simple ufunc: np.add
-            self.assertEqual(pickle_depickle(np.add), np.add)
+            self.assertEqual(pickle_depickle(np.add, protocol=self.protocol),
+                             np.add)
         else:  # skip if numpy is not available
             pass
 
         if spp:
             # custom ufunc: scipy.special.iv
-            self.assertEqual(pickle_depickle(spp.iv), spp.iv)
+            self.assertEqual(pickle_depickle(spp.iv, protocol=self.protocol),
+                             spp.iv)
         else:  # skip if scipy is not available
             pass
 
     def test_loads_namespace(self):
         obj = 1, 2, 3, 4
-        returned_obj = cloudpickle.loads(cloudpickle.dumps(obj))
+        returned_obj = cloudpickle.loads(cloudpickle.dumps(
+            obj, protocol=self.protocol))
         self.assertEqual(obj, returned_obj)
 
     def test_load_namespace(self):
         obj = 1, 2, 3, 4
-        bio = BytesIO()
+        bio = io.BytesIO()
         cloudpickle.dump(obj, bio)
         bio.seek(0)
         returned_obj = cloudpickle.load(bio)
@@ -406,7 +413,7 @@ class CloudPickleTest(unittest.TestCase):
             def f(self, x):
                 return x + 1
 
-        g = pickle_depickle(F.f)
+        g = pickle_depickle(F.f, protocol=self.protocol)
         self.assertEqual(g.__name__, F.f.__name__)
         if sys.version_info[0] < 3:
             self.assertEqual(g.im_class.__name__, F.f.im_class.__name__)
@@ -459,8 +466,7 @@ class CloudPickleTest(unittest.TestCase):
         '''
         exec(textwrap.dedent(code), mod.__dict__)
 
-        pickled_module_path = 'mod_f.pkl'
-
+        pickled_module_path = os.path.join(self.tmpdir, 'mod_f.pkl')
         child_process_script = '''
         import pickle
         from cloudpickle.cloudpickle import _dynamic_modules_globals
@@ -485,11 +491,11 @@ class CloudPickleTest(unittest.TestCase):
         '''
 
         child_process_script = child_process_script.format(
-                pickled_module_path=pickled_module_path)
+                pickled_module_path=_escape(pickled_module_path))
 
         try:
             with open(pickled_module_path, 'wb') as f:
-                cloudpickle.dump(mod.func, f)
+                cloudpickle.dump(mod.func, f, protocol=self.protocol)
 
             assert_run_python_script(textwrap.dedent(child_process_script))
 
@@ -501,7 +507,7 @@ class CloudPickleTest(unittest.TestCase):
         # correctly serialized. This notably checks that the globals are
         # accessible and that there is no issue with the builtins (see #211)
 
-        pickled_func_path = 'local_func_g.pkl'
+        pickled_func_path = os.path.join(self.tmpdir, 'local_func_g.pkl')
 
         child_process_script = '''
         import pickle
@@ -513,7 +519,7 @@ class CloudPickleTest(unittest.TestCase):
         '''
 
         child_process_script = child_process_script.format(
-                pickled_func_path=pickled_func_path)
+                pickled_func_path=_escape(pickled_func_path))
 
         try:
 
@@ -521,7 +527,7 @@ class CloudPickleTest(unittest.TestCase):
 
             g = make_local_function()
             with open(pickled_func_path, 'wb') as f:
-                cloudpickle.dump(g, f)
+                cloudpickle.dump(g, f, protocol=self.protocol)
 
             assert_run_python_script(textwrap.dedent(child_process_script))
 
@@ -544,8 +550,10 @@ class CloudPickleTest(unittest.TestCase):
         # the pickled dynamic module, and then re-pickle it under a new name.
         # Finally, it will create a child process that will load the re-pickled
         # dynamic module.
-        parent_process_module_file = 'dynamic_module_from_parent_process.pkl'
-        child_process_module_file = 'dynamic_module_from_child_process.pkl'
+        parent_process_module_file = os.path.join(
+            self.tmpdir, 'dynamic_module_from_parent_process.pkl')
+        child_process_module_file = os.path.join(
+            self.tmpdir, 'dynamic_module_from_child_process.pkl')
         child_process_script = '''
             import pickle
             import textwrap
@@ -560,7 +568,7 @@ class CloudPickleTest(unittest.TestCase):
                 mod = pickle.load(f)
 
             with open('{child_process_module_file}', 'wb') as f:
-                cloudpickle.dump(mod, f)
+                cloudpickle.dump(mod, f, protocol={protocol})
 
             assert_run_python_script(textwrap.dedent(child_of_child_process_script))
             '''
@@ -579,13 +587,14 @@ class CloudPickleTest(unittest.TestCase):
                 child_process_module_file=child_process_module_file)
 
         child_process_script = child_process_script.format(
-                parent_process_module_file=parent_process_module_file,
-                child_process_module_file=child_process_module_file,
-                child_of_child_process_script=child_of_child_process_script)
+            parent_process_module_file=_escape(parent_process_module_file),
+            child_process_module_file=_escape(child_process_module_file),
+            child_of_child_process_script=_escape(child_of_child_process_script),
+            protocol=self.protocol)
 
         try:
             with open(parent_process_module_file, 'wb') as fid:
-                cloudpickle.dump(mod, fid)
+                cloudpickle.dump(mod, fid, protocol=self.protocol)
 
             assert_run_python_script(textwrap.dedent(child_process_script))
 
@@ -606,7 +615,7 @@ class CloudPickleTest(unittest.TestCase):
         def my_small_function(x, y):
             return nested_function(x) + y
 
-        b = cloudpickle.dumps(my_small_function)
+        b = cloudpickle.dumps(my_small_function, protocol=self.protocol)
 
         # Make sure that the pickle byte string only includes the definition
         # of my_small_function and its dependency nested_function while
@@ -681,7 +690,7 @@ class CloudPickleTest(unittest.TestCase):
             res = yield f(0.01, y)
             raise gen.Return(res + 1)
 
-        data = cloudpickle.dumps([g, g])
+        data = cloudpickle.dumps([g, g], protocol=self.protocol)
         f = g = None
         g2, g3 = pickle.loads(data)
         self.assertTrue(g2 is g3)
@@ -709,7 +718,7 @@ class CloudPickleTest(unittest.TestCase):
         exec(textwrap.dedent(code), d, d)
         f = d['f']
         res = f()
-        data = cloudpickle.dumps([f, f])
+        data = cloudpickle.dumps([f, f], protocol=self.protocol)
         d = f = None
         f2, f3 = pickle.loads(data)
         self.assertTrue(f2 is f3)
@@ -727,7 +736,7 @@ class CloudPickleTest(unittest.TestCase):
         def example():
             x = xml.etree.ElementTree.Comment # potential AttributeError
 
-        s = cloudpickle.dumps(example)
+        s = cloudpickle.dumps(example, protocol=self.protocol)
 
         # refresh the environment, i.e., unimport the dependency
         del xml
@@ -748,7 +757,7 @@ class CloudPickleTest(unittest.TestCase):
             return example
         example = scope()
 
-        s = cloudpickle.dumps(example)
+        s = cloudpickle.dumps(example, protocol=self.protocol)
 
         # refresh the environment (unimport dependency)
         for item in list(sys.modules):
@@ -761,16 +770,14 @@ class CloudPickleTest(unittest.TestCase):
     def test_multiprocess(self):
         # running a function pickled by another process (a la dask.distributed)
         def scope():
-            import curses.textpad
             def example():
                 x = xml.etree.ElementTree.Comment
-                x = curses.textpad.Textbox
             return example
         global xml
         import xml.etree.ElementTree
         example = scope()
 
-        s = cloudpickle.dumps(example)
+        s = cloudpickle.dumps(example, protocol=self.protocol)
 
         # choose "subprocess" rather than "multiprocessing" because the latter
         # library uses fork to preserve the parent environment.
@@ -785,15 +792,15 @@ class CloudPickleTest(unittest.TestCase):
         # (unlike test_submodule)
         global etree
         def scope():
-            import curses.textpad as foobar
+            import xml.etree as foobar
             def example():
                 x = etree.Comment
-                x = foobar.Textbox
+                x = foobar.ElementTree
             return example
         example = scope()
         import xml.etree.ElementTree as etree
 
-        s = cloudpickle.dumps(example)
+        s = cloudpickle.dumps(example, protocol=self.protocol)
 
         command = ("import pickle, base64; "
                    "pickle.loads(base64.b32decode('" +
@@ -822,12 +829,12 @@ class CloudPickleTest(unittest.TestCase):
         dumped = cloudpickle.dumps(logger)
 
         code = """if 1:
-            import cloudpickle, logging
+            import base64, cloudpickle, logging
 
             logging.basicConfig(level=logging.INFO)
-            logger = cloudpickle.loads(%(dumped)r)
+            logger = cloudpickle.loads(base64.b32decode(b'{}'))
             logger.info('hello')
-            """ % locals()
+            """.format(base64.b32encode(dumped).decode('ascii'))
         proc = subprocess.Popen([sys.executable, "-c", code],
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.STDOUT)
@@ -944,7 +951,7 @@ class CloudPickleTest(unittest.TestCase):
         # serializable.
         from cloudpickle import CloudPickler
         CloudPickler.dispatch[type(py.builtin)] = CloudPickler.save_module
-        g = cloudpickle.loads(cloudpickle.dumps(f))
+        g = cloudpickle.loads(cloudpickle.dumps(f, protocol=self.protocol))
 
         result = g()
         self.assertEqual(1, result)
@@ -1139,6 +1146,11 @@ class CloudPickleTest(unittest.TestCase):
         '''
         exec(textwrap.dedent(code), mod.__dict__)
 
+        with_initial_globals_file = os.path.join(
+            self.tmpdir, 'function_with_initial_globals.pkl')
+        with_modified_globals_file = os.path.join(
+            self.tmpdir, 'function_with_modified_globals.pkl')
+
         try:
             # Simple sanity check on the function's output
             assert mod.func_defined_in_dynamic_module() == "initial value"
@@ -1153,7 +1165,7 @@ class CloudPickleTest(unittest.TestCase):
             # when unpickling the second function whatever the global variable
             # GLOBAL_STATE's value at the time of pickling.
 
-            with open('function_with_initial_globals.pkl', 'wb') as f:
+            with open(with_initial_globals_file, 'wb') as f:
                 cloudpickle.dump(mod.func_defined_in_dynamic_module, f)
 
             # Change the mod's global variable
@@ -1162,13 +1174,14 @@ class CloudPickleTest(unittest.TestCase):
             # At this point, mod.func_defined_in_dynamic_module()
             # returns the updated value. Let's pickle it again.
             assert mod.func_defined_in_dynamic_module() == 'changed value'
-            with open('function_with_modified_globals.pkl', 'wb') as f:
-                cloudpickle.dump(mod.func_defined_in_dynamic_module, f)
+            with open(with_modified_globals_file, 'wb') as f:
+                cloudpickle.dump(mod.func_defined_in_dynamic_module, f,
+                                 protocol=self.protocol)
 
             child_process_code = """
                 import pickle
 
-                with open('function_with_initial_globals.pkl','rb') as f:
+                with open({with_initial_globals_file!r},'rb') as f:
                     func_with_initial_globals = pickle.load(f)
 
                 # At this point, a module called 'mod' should exist in
@@ -1179,7 +1192,7 @@ class CloudPickleTest(unittest.TestCase):
 
                 # Load a function with initial global variable that was
                 # pickled after a change in the global variable
-                with open('function_with_modified_globals.pkl','rb') as f:
+                with open({with_initial_globals_file!r},'rb') as f:
                     func_with_modified_globals = pickle.load(f)
 
                 # assert the this unpickling did not modify the value of
@@ -1191,16 +1204,18 @@ class CloudPickleTest(unittest.TestCase):
                 assert func_with_initial_globals('new value') == 'new value'
                 assert func_with_modified_globals() == 'new value'
 
-                with open('function_with_initial_globals.pkl','rb') as f:
+                with open({with_initial_globals_file!r},'rb') as f:
                     func_with_initial_globals = pickle.load(f)
                 assert func_with_initial_globals() == 'new value'
                 assert func_with_modified_globals() == 'new value'
-            """
+            """.format(
+                with_initial_globals_file=_escape(with_initial_globals_file),
+                with_modified_globals_file=_escape(with_modified_globals_file))
             assert_run_python_script(textwrap.dedent(child_process_code))
 
         finally:
-            os.unlink('function_with_initial_globals.pkl')
-            os.unlink('function_with_modified_globals.pkl')
+            os.unlink(with_initial_globals_file)
+            os.unlink(with_modified_globals_file)
 
     @pytest.mark.skipif(sys.version_info >= (3, 0),
                         reason="hardcoded pickle bytes for 2.7")
@@ -1233,11 +1248,11 @@ class CloudPickleTest(unittest.TestCase):
         for exc_type in [Exception, ValueError, TypeError, RuntimeError]:
             obj = RaiserOnPickle(exc_type("foo"))
             with pytest.raises((exc_type, pickle.PicklingError)):
-                cloudpickle.dumps(obj)
+                cloudpickle.dumps(obj, protocol=self.protocol)
 
     def test_unhashable_function(self):
         d = {'a': 1}
-        depickled_method = pickle_depickle(d.get)
+        depickled_method = pickle_depickle(d.get, protocol=self.protocol)
         self.assertEqual(depickled_method('a'), 1)
         self.assertEqual(depickled_method('b'), None)
 
@@ -1265,7 +1280,7 @@ class CloudPickleTest(unittest.TestCase):
         def g():
             f()
 
-        f2 = pickle_depickle(g)
+        f2 = pickle_depickle(g, protocol=self.protocol)
 
         self.assertEqual(f2.__name__, f.__name__)
 
@@ -1280,7 +1295,7 @@ class CloudPickleTest(unittest.TestCase):
         def g():
             f()
 
-        f2 = pickle_depickle(g)
+        f2 = pickle_depickle(g, protocol=self.protocol)
 
         self.assertEqual(f2.__doc__, f.__doc__)
 
@@ -1299,10 +1314,27 @@ class CloudPickleTest(unittest.TestCase):
         def g(x):
             f(x)
 
-        f2 = pickle_depickle(g)
+        f2 = pickle_depickle(g, protocol=self.protocol)
 
         self.assertEqual(f2.__annotations__, f.__annotations__)
 
+    def test_instance_with_slots(self):
+        for slots in [["registered_attribute"], "registered_attribute"]:
+            class ClassWithSlots(object):
+                __slots__ = slots
+
+                def __init__(self):
+                    self.registered_attribute = 42
+
+            initial_obj = ClassWithSlots()
+            depickled_obj = pickle_depickle(
+                initial_obj, protocol=self.protocol)
+
+            for obj in [initial_obj, depickled_obj]:
+                self.assertEqual(obj.registered_attribute, 42)
+                with pytest.raises(AttributeError):
+                    obj.non_registered_attribute = 1
+
 
 class Protocol2CloudPickleTest(CloudPickleTest):
 
